# -*- coding: utf-8 -*-
"""
This module contains all functions, necessary for the inititation of an emipy project.
"""

import pandas as pd
import os

import requests
import zipfile
import io

import configparser
import urllib.request

path_sep = os.sep


def download_PollutionData(path, chunk_size=128):
    """
    Downloads the pollution data into given path.

    Parameters
    ----------
    path : String
        Path to the root of the project.
    chunk_size : TYPE, optional
        DESCRIPTION. The default is 128.

    Returns
    -------
    None.

    """
    directory = 'PollutionData'
    path = os.path.join(path, directory)
    if os.path.isdir(path) is False:
        os.mkdir(path)
    if not os.listdir(path):
        url = ('https://www.eea.europa.eu/data-and-maps/data/' +
               'member-states-reporting-art-7-under-the-european-pollutant' +
               '-release-and-transfer-register-e-prtr-regulation-23/european-' +
               'pollutant-release-and-transfer-register-e-prtr-data-base/eprtr_v9_csv.zip/at_download/file')
        r = requests.get(url, stream=True)
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(path)


def download_MapData(path, resolution=10, clear=False, chunk_size=128):
    """
    Download shp files to given path

    Parameters
    ----------
    path : String
        Path to the root of the project.
    resolution : int/list, optional
        Defines the resolution, that the downloaded maps have. Selectable are 1,3,10,20,60. The default is 10.
    clear : Boolean
        If put on True, the function clears the MappingData Directory, before downloading the data., The default is False.
    chunk_size : TYPE, optional
        DESCRIPTION. The default is 128.

    Returns
    -------
    None.

    """
    oldcwd = os.getcwd()
    directory = 'MappingData'
    path = os.path.join(path, directory)

    try:
        if os.path.isdir(path) is False:
            os.mkdir(path)

        if clear:
            os.chdir(path)
            for item in os.listdir(path):
                file_name = os.path.abspath(item)
                os.remove(file_name)

        urldict = {
            202101: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2021-01m.shp.zip',
            202103: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2021-03m.shp.zip',
            202110: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2021-10m.shp.zip',
            202120: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2021-20m.shp.zip',
            202160: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2021-60m.shp.zip',

            201601: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2016-01m.shp.zip',
            201603: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2016-03m.shp.zip',
            201610: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2016-10m.shp.zip',
            201620: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2016-20m.shp.zip',
            201660: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2016-60m.shp.zip',

            201301: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2013-01m.shp.zip',
            201303: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2013-03m.shp.zip',
            201310: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2013-10m.shp.zip',
            201320: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2013-20m.shp.zip',
            201360: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2013-60m.shp.zip',

            201001: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2010-01m.shp.zip',
            201003: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2010-03m.shp.zip',
            201010: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2010-10m.shp.zip',
            201020: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2010-20m.shp.zip',
            201060: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2010-60m.shp.zip',

            200601: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2006-01m.shp.zip',
            200603: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2006-03m.shp.zip',
            200610: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2006-10m.shp.zip',
            200620: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2006-20m.shp.zip',
            200660: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2006-60m.shp.zip',

            200301: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2003-01m.shp.zip',
            200303: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2003-03m.shp.zip',
            200310: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2003-10m.shp.zip',
            200320: 'http://gisco-services.ec.europa.eu/distribution/v2/nuts/download/ref-nuts-2003-20m.shp.zip',
        }

        publist = [2021, 2016, 2013, 2010, 2006, 2003]
        urllist = []
        for items in publist:
            if isinstance(resolution, list):
                for res in resolution:
                    if len(str(res)) == 1:
                        urlkey = int(str(items) + str(0) + str(res))
                    else:
                        urlkey = int(str(items) + str(res))
                    urllist.append(urldict.get(urlkey))
            else:
                if len(str(resolution)) == 1:
                    urlkey = int(str(items) + str(0) + str(resolution))
                else:
                    urlkey = int(str(items) + str(resolution))
                urllist.append(urldict.get(urlkey))

        for items in urllist:
            if items is None:
                continue
            r = requests.get(items, stream=True)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            z.extractall(path)
            extension = '.zip'
            os.chdir(path)
            for item in os.listdir(path):
                if item.endswith(extension):
                    file_name = os.path.abspath(item)
                    zip_ref = zipfile.ZipFile(file_name)
                    zip_ref.extractall(path)
                    zip_ref.close()
                    os.remove(file_name)
    finally:
        os.chdir(oldcwd)


def download_NACE_TransitionTables(path):
    """
    Creates, if necessary, the folder TransitionData in the given path and downloads NACE transition tables from Eurostat to save them in the folder TransitionData.

    Parameters
    ----------
    path : String
        Path to the root of the project.

    Returns
    -------
    None.

    """
    directory = 'TransitionData'
    path = os.path.join(path, directory)
    if os.path.isdir(path) is False:
        os.mkdir(path)
    if not os.listdir(path):
        url1 = (r'https://ec.europa.eu/eurostat/documents/1965800/1978760/' +
                '52249E90762B69F8E0440003BA9322F9.xls/2314229f-7606-4230-96d4-da18f045b692')
        url2 = (r'https://ec.europa.eu/eurostat/documents/1965800/1978760/' +
                'Copy+of+5225C5EEBF016050E0440003BA9322F9.xls/6e1aec88-f15b-4d0f-bccc-4ee505c7f810')
        urllib.request.urlretrieve(url1, os.path.join(path, 'Correspondance+table+NACERev2-NACERev1_1+table+format.xls'))
        urllib.request.urlretrieve(url2, os.path.join(path, 'Correspondance+table+NACERev1_1-NACERev2+table+format.xls'))


def pickle_RawData(path, force_rerun=False):
    """
    loads files of interest, converts them into .pkl file and saves them in the same path.

    Parameters
    ----------
    path : String
        Path to the root of the project.
    force_rerun : Boolean, optional
        If true, the function executes the routine even if the destination folder already contains corresponding files.. The default is False.

    Returns
    -------
    None.

    """

    # POLLUTANTRELEASEANDTRANSFERREPORT
    if (os.path.isfile(os.path.join(path, 'PollutionData', 'pratr.pkl')) is False) or force_rerun:
        pratr = pd.read_csv(os.path.join(path, 'PollutionData', 'dbo.PUBLISH_POLLUTANTRELEASEANDTRANSFERREPORT.csv'))
        pratr.to_pickle(os.path.join(path, 'PollutionData', 'pratr.pkl'))

    # FACILITYREPORT
    if (os.path.isfile(os.path.join(path, 'PollutionData', 'fr.pkl')) is False) or force_rerun:
        fr = pd.read_csv(os.path.join(path, 'PollutionData', 'dbo.PUBLISH_FACILITYREPORT.csv'), encoding='latin-1', low_memory=False)
        fr.to_pickle(os.path.join(path, 'PollutionData', 'fr.pkl'))

    # POLLUTANTRELEASE
    if (os.path.isfile(os.path.join(path, 'PollutionData', 'pr.pkl')) is False) or force_rerun:
        pr = pd.read_csv(os.path.join(path, 'PollutionData', 'dbo.PUBLISH_POLLUTANTRELEASE.csv'), low_memory=False)
        pr.to_pickle(os.path.join(path, 'PollutionData', 'pr.pkl'))

    return None


def merge_PollutionData(path, force_rerun=False):
    """
    Inserts tables with different data into each other.

    Parameters
    ----------
    path : String
        Path to the root of the project.
    force_rerun : Boolean, optional
        If true, the function executes the routine even if the destination folder already contains corresponding files.. The default is False.

    Returns
    -------
    None.

    """
    if (os.path.isfile(os.path.join(path, 'PollutionData', 'db.pkl')) is False) or force_rerun:
        try:
            fr = pd.read_pickle(os.path.join(path, 'PollutionData', 'fr.pkl'))
            pr = pd.read_pickle(os.path.join(path, 'PollutionData', 'pr.pkl'))
            pratr = pd.read_pickle(os.path.join(path, 'PollutionData', 'pratr.pkl'))
        except FileNotFoundError:
            print('Error. Run function pickle_RawData')
        # speed difference for variing merge-order?? Table length differs,
        # merge smart enough to add multiple one row to multiple?
        # problematic to merge by multiple coloum names?
        # Some data have no PostalCode, wrong fomrated postal code or not actual PostalCode
        db01 = pd.merge(fr, pratr, on=['PollutantReleaseAndTransferReportID', 'CountryName', 'CountryCode'])
        db02 = pd.merge(db01, pr, on=['FacilityReportID', 'ConfidentialIndicator', 'ConfidentialityReasonCode',
                                      'ConfidentialityReasonName'])
        db02.to_pickle(os.path.join(path, 'PollutionData', 'db.pkl'))
    return None


def generate_PollutionData_2(path):
    """
    Generates the new data set. For this, the function downloads the .pkl file from the emipy repository, decompresses it and renames the columns.

    Parameters
    ----------
    path : String
        Path to the root of the project.

    Returns
    -------
    data2 : DataFrame
        The new data base in the format, that enables the emipy functions to act on it.

    """
    directory = 'PollutionData'
    path = os.path.join(path, directory)
    if os.path.isdir(path) is False:
        os.mkdir(path)
    if 'emipy_newdb.pkl' not in os.listdir(path):
        url = ('https://gitlab-public.fz-juelich.de/network-science-group/emipy/'
               + '-/raw/master/additionaldata/emipy_newdb.pkl?inline=false')
        urllib.request.urlretrieve(url, os.path.join(path, 'emipy_newdb.pkl'))
    if 'dbnew.pkl' not in os.listdir(path):
        data = pd.read_pickle(os.path.join(path, 'emipy_newdb.pkl'), compression='xz')
        columndict = {
            'reportingYear': 'ReportingYear',
            'Facility_INSPIRE_ID': 'FacilityReportID',
            'parentCompanyName': 'ParentCompanyName',
            'nameOfFeature': 'FacilityName',
            'mainActivityCode': 'MainIAActivityCode',
            'mainActivityName': 'MainIAActivityName',
            'pointGeometryLon': 'Long',
            'pointGeometryLat': 'Lat',
            'streetName': 'StreetName',
            'buildingNumber': 'BuildingNumber',
            'postalCode': 'PostalCode',
            'city': 'City',
            'countryCode': 'CountryCode',
            'pollutantCode': 'PollutantCode',
            'pollutantName': 'PollutantName',
            'medium': 'ReleaseMediumCode',
            'totalPollutantQuantityKg': 'TotalQuantity',
            'AccidentalPollutantQuantityKG': 'AccidentalQuantity',
            'methodCode': 'MethodBasisCode',
            'methodName': 'MethodBasisName',
            'NUTSRegionSourceCode': 'NUTSRegionGeoCode',
            'NUTSRegionSourceName': 'NUTSRegionGeoName',
            'NACEMainEconomicActivityCode': 'NACEMainEconomicActivityCode',
            'NACEMainEconomicActivityName': 'NACEMainEconomicActivityName'
        }
        data2 = data.rename(columndict, axis='columns')

        data2.loc[:, 'TotalQuantity'] = data2.loc[:, 'TotalQuantity'].replace(',', '.', regex=True).astype(float)
        data2.loc[:, 'AccidentalQuantity'] = data2.loc[:, 'AccidentalQuantity'].replace(',', '.', regex=True).astype(float)
        data2.loc[:, 'Long'] = data2.loc[:, 'Long'].replace(',', '.', regex=True).astype(float)
        data2.loc[:, 'Lat'] = data2.loc[:, 'Lat'].replace(',', '.', regex=True).astype(float)
        data2.loc[:, 'NACEMainEconomicActivityCode'] = data2.loc[:, 'NACEMainEconomicActivityCode'].astype(str).apply(lambda x: '0' + x if x.find('.') == 1 else x).apply(lambda x: x + '0' if len(x) == 4 else x)

        data2.to_pickle(os.path.join(path, 'dbnew.pkl'))
    os.remove(os.path.join(path, 'emipy_newdb.pkl'))


def get_RootPath():
    """
    Returns the current root path, stored in the config file.

    Returns
    -------
    path : String
        Current path to the root of the project, stored in the config file.

    """
    config = configparser.ConfigParser()
    config.read(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'configuration',
                             'configuration.ini'))
    path = config['PATH']['path']
    return path


def change_RootPath(path):
    """
    Changes the path of the root to the project in the configuration.ini file.

    Parameters
    ----------
    path : String
        Path to the project, which is to access.

    Returns
    -------
    None.

    """
    config = configparser.ConfigParser()
    print(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'configuration',
                       'configuration.ini'))
    config.read(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'configuration',
                             'configuration.ini'))
    config.set('PATH', 'path', path)
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'configuration',
                           'configuration.ini'), 'w') as configfile:
        config.write(configfile)


def init_emipy_project(path, resolution=10, force_rerun=False):
    """
    Executes the initiation of a new project. Downloads all needed data to the given path and merges data of interest.

    Parameters
    ----------
    path : String
        Path to root of project.
    force_rerun : Boolean, optional
        Forces the programm to rerun the merging routine, if True. The default is False.
    resolution : int/list, optional
        Defines the resolution, that the downloaded maps have. Selectable are 1,3,10,20,60. The default is 10.

    Returns
    -------
    None.

    """
    if path is None:
        print('A path to the root of the project is needed to initialize the project.')
        return

    change_RootPath(path)
    print('Root path successfully changed.')
    download_PollutionData(path=path)
    print('Pollution data successfully downloaded and extracted.')
    download_MapData(path=path, resolution=resolution)
    print('Map data successfully downloaded and extractet.')
    download_NACE_TransitionTables(path)
    print('Transition tables successfully downloaded.')
    pickle_RawData(path=path, force_rerun=force_rerun)
    merge_PollutionData(path=path, force_rerun=force_rerun)
    print('Data merged and pickled.')
    generate_PollutionData_2(path)
    print('New data base with data from 2017-2019 is generated.')
    directory = 'ExportData'
    path = os.path.join(path, directory)
    if os.path.isdir(path) is False:
        os.mkdir(path)
    print('The initialisation process is completed.')

    return



# -*- coding: utf-8 -*-
"""
This module contains all functions to visualize the data set.
"""

import pandas as pd
import os
import matplotlib.pyplot as plt
import numpy as np
import geopandas as gpd
import configparser
import sys


def get_PollutantVolume(db, FirstOrder=None, SecondOrder=None):
    """
    Sorts the input data table, to the named order parameters, which are all possible column names.

    Parameters
    ----------
    db : DataFrame
        input data table.
    FirstOrder : String, optional
        Name of column, the data are sorted in the first order. The default is None.
    SecondOrder : TYPE, optional
        Name of column, the data are sorted in the second order. The default is None.

    Returns
    -------
    data : DataFrame
        Data table, sorted to the announced order parameters.

    """
    if SecondOrder is None:
        if FirstOrder is None:
            d = {'Order': ['NoneGiven'], 'TotalQuantity': [db.TotalQuantity.sum()]}
            data = pd.DataFrame(data=d)
        else:
            data = db.groupby([FirstOrder]).sum().reset_index()
            data = data[[FirstOrder, 'TotalQuantity']]
    else:
        timer = 0
        for items in db[SecondOrder].unique():
            if timer == 0:
                timer = 1
                data = db[db[SecondOrder] == items].groupby([FirstOrder]).TotalQuantity.sum().reset_index()
                data = data.rename(columns={'TotalQuantity': items})
            else:
                itemdata = db[db[SecondOrder] == items].groupby([FirstOrder]).TotalQuantity.sum().reset_index()
                data = pd.merge(data, itemdata, on=[FirstOrder], how='outer')
                data = data.rename(columns={'TotalQuantity': items})
    return data


def get_PollutantVolumeRel(db, FirstOrder=None, SecondOrder=None, normtop=None, normtov=None):
    """
    Normalises the volume values to one specific value. This value is either the present max value of the returned data table or is specifed by normtop(osition) or normtov(alue).

    Parameters
    ----------
    db : DataFrame
        input data table.
    FirstOrder : String, optional
        Name of column, the data are sorted in the first order. The default is None.
    SecondOrder : String, optional
        Name of column, the data are sorted in the second order. The default is None.
    normtop : list, optional
        With this parameter you can choose a entry of your data table, that the entries should be normalised too. The first item of the list has to be one value of the FirstOrder. If SecondOrder is called, the second value has to be a value of the SecondOrder. The default is None.
    normtov : float, optional
        With this parameter you can define a value, that the PollutantVolume entries are normalised to. The default is None
    Returns
    -------
    data : DataFrame
        Data table sorted to the announced parameters. The values are normed to one specific max value. If normtop and normtov are both unequal None, no normalization is applied, since there is no concrete value, that can be normed to.

    """
    data = get_PollutantVolume(db, FirstOrder=FirstOrder, SecondOrder=SecondOrder)

    if (normtop is not None) and (normtov is not None):
        print('Multiple normto values given. Decide for either normtoposition or normtovalue. No normalisation applied!')
        return data

    if (normtop is None) and (normtov is None):
        maxvalue = np.nanmax(data.iloc[:, 1:].to_numpy())
    elif normtop is not None:
        if FirstOrder is None:
            maxvalue = data.iloc[0, 1]
        elif SecondOrder is None:
            maxvalue = data[data[FirstOrder] == normtop[0]].reset_index().loc[0, 'TotalQuantity']
        else:
            maxvalue = data[data[FirstOrder] == normtop[0]].reset_index().loc[0, normtop[1]]
    elif normtov is not None:
        maxvalue = normtov

    if pd.isnull(maxvalue):
        print('The determined maxvalue is of type Nan. Therefore, the normalization does not' +
              ' work and a DataFrame with NAN is returned.')

    data.iloc[:, 1:] = data.iloc[:, 1:] / maxvalue
    return data


def get_PollutantVolumeChange(db, FirstOrder=None, SecondOrder=None):
    """
    Derives the pollutant volume change to the previous year.

    Parameters
    ----------
    db : DataFrame
        the filtered input DataFrame.
    FirstOrder : String, optional
        Name of column, the data are sorted in the first order. The default is None.
    SecondOrder : String, optional
        Name of column, the data are sorted in the second order. The default is None.

    Returns
    -------
    data : DataFrame
        The change of TotalQuantity to the previous data entry

    """
    data = get_PollutantVolume(db, FirstOrder=FirstOrder, SecondOrder=SecondOrder)
    if SecondOrder is None:
        data = data.rename(columns={'TotalQuantity': 'TotalQuantityChange'})
    for items in data.columns:
        if items != FirstOrder:
            data[items] = data[items].diff()
    data = data.iloc[1:]
    return data


def get_ImpurityVolume(db, target, FirstOrder='FacilityReportID', ReleaseMediumName='Air', absolute=False, FacilityFocus=True, impurity=None, statistics=False):
    """
    Creates a table with the impurities of the target pollutant, sorted by the FirstOrder parameter. Putting the absolute parameter to True, gives absolute values instead of relative.

    Parameters
    ----------
    db : DataFrame
        Data to look for impurities.
    target : String
        Pollutant name of the pollutant, which is not seen as impurity.
    FirstOrder : String, optional
        Order to sort the impurities by. E.g. NACERegionGeoCode, FacilityReportID, NACEMainEconomicActivityCode. The default is 'FacilityReportID'.
    ReleaseMediumName : String, optional
        The release medium name in which the target is emitted and in which can be impurities. The default is 'Air'.
    absolute : Boolean, optional
        If this parameter is set on False, this function returns the impurity relative to the target pollutant emission. If it is set on True, the absolute emission value is returned. The default is False.
    FacilityFocus : Boolean, optional
        If this parameter is true, only the impurities in the facilities in which the target is emittet is taken in to consideration. If it is false, all data are taken into consideration. The default is True.
    impurity : String, optional
        With this parameter, you can specify the impurity pollutant you want to return. Otherwise, all present impurities are shown. The default is None.
    statistics : Boolean, optional
        If this argument is True, the statistics (determined by .describe()) of the output DataFrame are returned, instead of the usual impurity table. The default is False.

    Returns
    -------
    d2 : DataFrame
        Data table with the rows beeing the different present order values, and in the columns their respective emission of the target pollutant and the absolute emission of the impurities.
    d3 : DataFrame
        Data table with the rows beeing the different present order values, and in the columns their respective emission of the target pollutant and the relative emission of the impurities.

    """
    db = db[db.ReleaseMediumName == ReleaseMediumName]
    d1 = db[db.PollutantName == target]
    d2 = get_PollutantVolume(d1, FirstOrder=FirstOrder)
    d3 = get_PollutantVolume(d1, FirstOrder=FirstOrder)
    if FacilityFocus == True:
        ff = get_PollutantVolume(d1, FirstOrder='FacilityReportID').FacilityReportID.unique()
        db = db[db.FacilityReportID.isin(ff)]
    for items in np.delete(db.PollutantName.unique(), np.argwhere(db.PollutantName.unique() == target)):
        item = db[db.PollutantName == items].groupby([FirstOrder]).TotalQuantity.sum().reset_index()
        item = item[[FirstOrder, 'TotalQuantity']].rename(columns={'TotalQuantity': items})
        d2 = d2.merge(item, how='left', on=FirstOrder)
        d3[items] = d2.loc[:, items] / d2.loc[:, 'TotalQuantity']

    if impurity != None:
        d2 = d2[[FirstOrder, 'TotalQuantity', impurity]]
        d3 = d3[[FirstOrder, 'TotalQuantity', impurity]]
    
    if absolute:
        if statistics:
            return d2.describe()
        else:
            return d2
    else:
        if statistics:
            return d3.describe()
        else:
            return d3


def plot_PollutantVolume(db, FirstOrder=None, SecondOrder=None, stacked=False, *args, **kwargs):
    """

    Plots the filtered data set. The first order is the x-axis, the second order is a differentiation of the y-values.

    Parameters
    ----------
    db : DataFrame
        The data to be plotted.
    FirstOrder : String, optional
        Name of column, the data are sorted in the first order. The default is None.
    SecondOrder : String, optional
        Name of column, the data are sorted in the second order. The default is None.
    stacked : Boolean, optional
        Stacks the bars for second order. The default is False.
    *args : TYPE
        pandas.plot() input variables.
    **kwargs : TYPE
        pandas.plot() input variables.

    Returns
    -------
    ax : Axes
        Plot of the data in db, sorted by FirstOrder and SecondOrder if given.

    """
    data = get_PollutantVolume(db, FirstOrder=FirstOrder, SecondOrder=SecondOrder)
    if SecondOrder is None:
        ax = data.plot(x=FirstOrder, y='TotalQuantity', kind='bar', *args, **kwargs)
    else:
        if stacked is True:
            ax = data.plot.bar(x=FirstOrder, stacked=True, *args, **kwargs)
        else:
            ax = data.plot.bar(x=FirstOrder, *args, **kwargs)

    return ax


def plot_PollutantVolumeRel(db, FirstOrder=None, SecondOrder=None, normtop=None, normtov=None, stacked=False, *args, **kwargs):
    """
    Plots the normed pollutant volume of the data set, The first order is the x-axis, the second order is a differentiation of the y-values.

    Parameters
    ----------
    db : DataFrame
        The data to be plotted.
    FirstOrder : String, optional
        Name of column, the data are sorted in the first order. The default is None.
    SecondOrder : String, optional
        Name of column, the data are sorted in the second order.. The default is None.
    normtop : list, optional
        With this parameter you can choose a entry of your data table, that the entries should be normalised too. The first item of the list has to be one value of the FirstOrder. If SecondOrder is called, the second value has to be a value of the SecondOrder. The default is None.
    normtov : float, optional
        With this parameter you can define a value, that the PollutantVolume entries are normalised to. The default is None.
    stacked : Boolean, optional
        Stacks the bars for second order. The default is False.
    *args : TYPE
        pandas.plot() input variables.
    **kwargs : TYPE
        pandas.plot() input variables.

    Returns
    -------
    ax : Axes
        Plot of the data in db, sorted by FirstOrder and SecondOrder if given.

    """
    data = get_PollutantVolumeRel(db, FirstOrder=FirstOrder, SecondOrder=SecondOrder, normtop=normtop, normtov=normtov)
    if SecondOrder is None:
        ax = data.plot(x=FirstOrder, y='TotalQuantity', kind='bar', *args, **kwargs)
    else:
        if stacked is True:
            ax = data.plot.bar(x=FirstOrder, stacked=True, *args, **kwargs)
        else:
            ax = data.plot.bar(x=FirstOrder, *args, **kwargs)
    return ax


def plot_PollutantVolumeChange(db, FirstOrder=None, SecondOrder=None, stacked=False, *args, **kwargs):
    """
    Plots the volume change of the data set. The first order is the x-axis, the second order
     is a differentiation of the y-values.

    Parameters
    ----------
    db : DataFrame
        The data to be plotted.
    FirstOrder : String, optional
        Name of column, the data are sorted in the first order. The default is None.
    SecondOrder : String, optional
        Name of column, the data are sorted in the second order.. The default is None.
    stacked : Boolean, optional
        Stacks the bars for second order. The default is False.
    *args : TYPE
        pandas.plot() input variables.
    **kwargs : TYPE
        pandas.plot() input variables.

    Returns
    -------
    ax : Axes
        Plot of the data in db, sorted by FirstOrder and SecondOrder if given.

    """
    data = get_PollutantVolumeChange(db, FirstOrder=FirstOrder, SecondOrder=SecondOrder)
    if SecondOrder is None:
        ax = data.plot(x=FirstOrder, y='TotalQuantityChange', kind='bar', *args, **kwargs)
    else:
        if stacked is True:
            ax = data.plot.bar(x=FirstOrder, stacked=True, *args, **kwargs)
        else:
            ax = data.plot.bar(x=FirstOrder, *args, **kwargs)
    return ax


def plot_ImpurityVolume(db, target, impurity, FirstOrder='FacilityReportID', ReleaseMediumName='Air', absolute=False, FacilityFocus=True, statistics=False, PlotNA=True, *args, **kwargs):
    """
    Plots the impurities for the different FirstOrder values or the statistics of the entries.

    Parameters
    ----------
    db : DataFrame
        The data to be plotted.
    target : String
        The target pollutant which is impured.
    impurity : String
        The impurity which is to be analyzed.
    FirstOrder : String, optional
        Name of column, the data are sorted in the first order. The default is 'FacilityReportID'.
    ReleaseMediumName : TYPE, optional
        The release medium name in which the target is emitted and in which can be impurities. The default is 'Air'.
    absolute : Boolean, optional
        If this parameter is set on False, this function returns the impurity relative to the target pollutant emission. If it is set on True, the absolute emission value is returned. The default is False.
    FacilityFocus : Boolean, optional
        If this parameter is true, only the impurities in the facilities in which the target is emittet is taken in to consideration. If it is false, all data are taken into consideration. The default is True.
    statistics : Boolean, optional
        If this parameter is True, the statistics of the data are plotted. If it is False, the actual values are plotted. The default is False.
    PlotNA : Boolean, optional
        This argument is a option for discarding the na values if plotting the impurities. The default is True.
    *args : TYPE
        pandas.plot() input variables.
    **kwargs : TYPE
        pandas.plot() input variables.

    Returns
    -------
    ax : Axes
        Plot of the impurities in db, or the statistics of these impurities.

    """
    data = get_ImpurityVolume(db=db, target=target, FirstOrder=FirstOrder, ReleaseMediumName=ReleaseMediumName, impurity=impurity, absolute=absolute, FacilityFocus=FacilityFocus, statistics=statistics)

    if statistics is True:
        ax = data.drop('count').plot(kind='bar', y=impurity, *args, **kwargs)
    else:
        if PlotNA:
            ax = data.drop('TotalQuantity', axis=1).plot(x=FirstOrder, y=impurity, kind='bar', *args, **kwargs)
        else:
            ax = data.drop('TotalQuantity', axis=1).dropna().plot(x=FirstOrder, y=impurity, kind='bar', *args, **kwargs)

    return ax


def get_mb_borders(mb):
    """
    Generates a list with the borders of the objects of a GeoDataFrame.

    Parameters
    ----------
    mb : GeoDataFrame
        Table of geo objects which over all borders are wanted.

    Returns
    -------
    borders : List
        The x,y min/max values.

    """
    foo = mb.bounds
    borders = (foo.minx.min(), foo.miny.min(), foo.maxx.max(), foo.maxy.max())
    return list(borders)


def exclude_DataOutsideBorders(borders, gdf):
    """
    seperates data, that are inside and outside given borders

    Parameters
    ----------
    borders : list
        x,y min/max.
    gdf : GeoDataFrame
        GeoDataFrame that is to process.

    Returns
    -------
    gdft : DataFrame
        GeoDataFrame with data inside the borders.
    gdff : DataFrame
        GeoDataFrame with data outside the borders.

    """
    gdft = gdf
    gdff = gpd.GeoDataFrame(columns=gdf.columns)
    gdff['geometry'] = ""
    for i in range(len(gdf)):
        if (gdf.geometry.iloc[i].x < borders[0]) or (gdf.geometry.iloc[i].x > borders[2]) or (gdf.geometry.iloc[i].y < borders[1]) or (gdf.geometry.iloc[i].y > borders[3]):
            gdff = gdff.append(gdf.iloc[i])
            gdft = gdft.drop([i], axis=0)
    return gdft, gdff


def add_MarkerSize(gdf, MaxMarker):
    """
    adds column MarkerSize to GeoDataFrame. If MaxMarker=0, all markers have size 1. Else, they are normalized to max value and multiplied by value of MaxMarker.

    Parameters
    ----------
    gdf : GeoDataFrame
        GeoDataFrame, which gets additional column.
    MaxMarker : Int
        defines the marker size of the biggest marker. If 0, all markers have same size.

    Returns
    -------
    gdf : GeoDataFrame
        GeoDataFrame with added column 'MarkerSize'.

    """
    gdf['MarkerSize'] = ""
    markernorm = gdf.TotalQuantity.max()
    if MaxMarker == 0:
        gdf['MarkerSize'] = 1
    else:
        gdf['MarkerSize'] = gdf['TotalQuantity'] / markernorm * MaxMarker
    return gdf


def create_GDFWithRightProj(dfgdf, OutProj=None):
    """
    Converts DataFrame into GeoDataFrame and changes the projection if new projection is given as input.

    Parameters
    ----------
    dfgdf : DataFrame/GeoDataFrame
        Data that is about to be converted into a GeoDataFrame and experience a projection change if wanted.
    OutProj : String, optional
        Target projection of the geometry of the data. The default is None.

    Returns
    -------
    gdf : GeoDataFrame
        Data stored as GeoDataFrame and with eventually changed geometry CRS.

    """
    if isinstance(dfgdf, pd.DataFrame):
        gdf = gpd.GeoDataFrame(dfgdf, geometry=gpd.points_from_xy(dfgdf.Long, dfgdf.Lat), crs='EPSG:4326').reset_index(drop=True)
    elif isinstance(dfgdf, gpd.GeoDataFrame):
        if dfgdf.crs is None:
            gdf = gpd.GeoDataFrame(dfgdf, geometry=gpd.points_from_xy(dfgdf.Long, dfgdf.Lat), crs='EPSG:4326').reset_index(drop=True)
    if OutProj != None:
        gdf = change_proj(gdf, OutProj=OutProj)
    return(gdf)


def change_proj(gdf, OutProj=None):
    """
    Changes The projection of the input GeoDataFrame to the projection defined with OutProj. If no CRS is given for the geometry, the function tries to recover information from gdf.

    Parameters
    ----------
    gdf : GeoDataFrame
        Data that CRS is to be changed.
    OutProj : Datatype, optional
        Code for target output projection. See http://pyproj4.github.io/pyproj/stable/api/crs/crs.html#pyproj.crs.CRS.from_user_input for input possibilities. The default is None.

    Returns
    -------
    gdf : GeoDataFrame
        Data with new projection in the geometry.

    """
    if OutProj is None:
        sys.exit('InputError: For the change of projection is a target projection required.')

    if gdf.crs is None:
        if 'Long' not in gdf.columns or 'Lat' not in gdf.columns:
            sys.exit('InputError: No information about projection of geometry.' +
                     ' Define CRS or give coordinates as Long and Lat!')
        gdf = gpd.GeoDataFrame(gdf, geometry=gpd.points_from_xy(gdf.Long, gdf.Lat), crs='EPSG:4326').reset_index(drop=True)
    gdf = gdf.to_crs(crs=OutProj)

    return gdf


def map_PollutantSource(db, mb, category=None, MarkerSize=0, OutProj=None,
                        ReturnMarker=0, *args, **kwargs):
    """
    maps pollutant sources given by db on map given by mb.

    Parameters
    ----------
    db : DataFrame/GeoDataFrame
        Data table on pollutant sources.
    mb : DataFrame
        geo data table.
    category : String
        The column name of db, which gets new colors for every unique entry.
    MarkerSize : Int
        maximal size of the largest marker.
    OutProj : DataType
        Code for targeted output projection. See http://pyproj4.github.io/pyproj/stable/api/crs/crs.html#pyproj.crs.CRS.from_user_input for input possibilities. The default is None.
    ReturnMarker : Int
        If put on 1, the function returns a DataFrame with all data that are plotted. If put on 2, the function returns a DataFrame with all data  that are not plotted, because their coordinates are outside the geo borders.
    *args : TYPE
        Geopandas.plot() input arguments.
    **kwargs : TYPE
        Geopandas.plot() input arguments.

    Returns
    -------
    ax : Axes
        Plot with pollutant sources on map.
    gdfp : GeoDataFrame
        GeoDataFrame with all sources that are within geo borders and therefore plotted.
    gdfd : GeoDataFrame
        GeoDataFrame with all sources that are outside geo borders and therefore dropped.

    """
    # color selecting is bad.
    # Calling gdfp, gdfd requires 2 times performing the function, perhaps better way.
    ax = mb.plot(zorder=1, *args, **kwargs)
    colorlist = ['r', 'y', 'g', 'c', 'm', 'b']
    borders = get_mb_borders(mb)
    if category is None:
        gdf = create_GDFWithRightProj(db, OutProj=OutProj)
        gdfp = exclude_DataOutsideBorders(borders=borders, gdf=gdf)[0]
        gdfd = exclude_DataOutsideBorders(borders=borders, gdf=gdf)[1]
        gdfp = add_MarkerSize(gdfp, MaxMarker=MarkerSize)
        ax = gdfp.plot(color='r', zorder=1, markersize=gdfp['MarkerSize'], *args, **kwargs)

    else:
        for items in db[category].unique():
            if not colorlist:
                print('Running out of color')
                break
            color = colorlist[0]
            colorlist.remove(color)
            itemdata = db[db[category] == items].reset_index()
            gdf = create_GDFWithRightProj(itemdata, OutProj=OutProj)
            gdfp = exclude_DataOutsideBorders(borders=borders, gdf=gdf)[0]
            gdfd = exclude_DataOutsideBorders(borders=borders, gdf=gdf)[1]
            gdfp = add_MarkerSize(gdfp, MaxMarker=MarkerSize)
            ax = gdfp.plot(color=color, zorder=1, markersize=gdfp['MarkerSize'], *args, **kwargs)

    if gdfd.empty is False:
        print('Some data points are out of borders')
    else:
        print('All data points are within rectangular borders')

    if ReturnMarker == 0:
        return ax

    elif ReturnMarker == 1:
        return gdfp

    else:
        return gdfd


def map_PollutantRegions(db, mb, ReturnMarker=0, *args, **kwargs):
    """
    Visualizes the pollutant emission of regions with a color map. The classification of regions is selected with the choice of mb. If ReturnMarker is put on 1, the function returns a DataFrame with the plotted data. If ReturnMarker is put on 2, the function returns the DataFrame with Data that have no complementary NUTSID in the mapdata.

    Parameters
    ----------
    db : DataFrame
        Pollution data that are plotted.
    mb : TYPE
        Map data for plotting. The region selection corresponds to the selection of mb.
    ReturnMarker : int
        If it has the value 0, the function returns the plot. If put on 1, the function returns a DataFrame with all data that are plotted. If put on 2, the function returns a DataFrame with all data that are not plotted, because their NUTS_ID is not present in the mapdata.
    *args : TYPE
        Geopandas.plot() input arguments.
    **kwargs : TYPE
        Geopandas.plot() input arguments.

    Returns
    -------
    ax : Axes
        Axes with colormap of the pollution emission.
    dbp : DataFrame
        Data that are plotted
    dbna : DataFrame
        Data that are not plotted, because the NUTS_ID is not present in the mapdata.

    """

    NUTSlvl = mb.LEVL_CODE.unique()
    if len(NUTSlvl) != 1:
        print('There are multiple NUTS-Levels present in the map data input.' +
              'This function can not seperate the data in the required way. No Output!')
        return None
    NUTSlvl = NUTSlvl[0]

    dbpremerge = pd.DataFrame(columns=['NUTS_ID', 'TotalQuantity'])
    if NUTSlvl <= 2:
        db01 = get_PollutantVolume(db, FirstOrder='NUTSRegionGeoCode')
        for item in mb.NUTS_ID.unique():
            itemdata = db01[db01.NUTSRegionGeoCode.str.startswith(item)]
            itemvalue = itemdata.TotalQuantity.sum()
            dbpremerge = dbpremerge.append({'NUTS_ID': item, 'TotalQuantity': itemvalue}, ignore_index=True)
    elif NUTSlvl == 3:
        print('The NUTS-Level of the map data is to high.' +
              ' The geospatial resolution of the emission data is not high' +
              ' enough to differentiate on level 3 regions. Use NUTS-LVL smaller or equal to 2.')
        return None

    db02 = pd.merge(mb, dbpremerge, how='left', on=['NUTS_ID'])
    ax = db02.plot(column='TotalQuantity', *args, **kwargs)

    presentNUTS_IDs = tuple(mb.NUTS_ID.tolist())
    dbp = db01[db01.NUTSRegionGeoCode.str.startswith(presentNUTS_IDs)]
    dbna = db01[~db01.NUTSRegionGeoCode.str.startswith(presentNUTS_IDs)]

    if ReturnMarker == 0:
        return ax

    elif ReturnMarker == 1:
        return dbp

    else:
        return dbna


def export_fig(fig, path=None, filename=None, **kwargs):
    """
    Exports the choosen figure to a given path or to the export folder of the project if no path is given.

    Parameters
    ----------
    fig : figure
        The figure that is to export.
    path : String, optional
        Path under which the file is stored. The filename has to be included. The default is None.
    filename : String, optional
        Filename under which the figure is stored in the Export folder of the project. The default is None.
    **kwargs : TYPE
        Matplotlib.savefig() input arguments.

    Returns
    -------
    None.

    """
    if path is None and filename is None:
        print('A filename is required.')
        return None

    elif path is None:
        config = configparser.ConfigParser()
        config.read(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                 'configuration', 'configuration.ini'))
        path = config['PATH']['path']
        path = os.path.join(os.path.join(path, 'ExportData'), filename)
    elif path is not None and filename is None:
        path = os.path.join(path, filename)

    if path.endswith(r'{}'.format(os.sep)):
        print('The file name or path must end with a file type extension')
        return None

    fig.savefig(path, **kwargs)

    return None



# -*- coding: utf-8 -*-
"""
Created on Sun May 17 10:25:20 2020

@author: f-ove
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="emipy",
    version="0.1.0",
    author="Florian Overberg, Philipp C. Böttcher, Simon Morgenthaler",
    author_email="p.boettcher@fz-juelich.de",
    description="Python package for emission data analysis based on the E-PRTR database.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://jugit.fz-juelich.de/network-science-group/emipy",
    packages=setuptools.find_packages(exclude=('tests',)),
    package_data={"emipy": ["configuration/*.ini"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        'Documentation': 'https://emipy.readthedocs.io/en/latest/',
        'Source': 'https://jugit.fz-juelich.de/network-science-group/emipy'
    },
    python_requires='>=3.6',
    install_requires=open('requirements.txt').read().splitlines(),


)
